class EventError(Exception):
    """Standard Event Exception"""


class Cancel(EventError):
    """Raised when an event is canceled"""


class StartError(EventError):
    """Raised when there is an error with starting the event manager"""



from functools import wraps

_monitoring = {}


def thread_monitor(f):
    """
    A decorator to bse used in monitoring functions that have any possibility of causing a deadlock

    Default implementation is under ConcurrentWorkLimiter and ResourceManager as both have
    functionality where threads can interact indirectly

    Debug can be used through this by simply printing the dictionary and it will display how many of each
    monitored function is running at any point in time

    Args:
        :param f: a function which will be added to the monitoring dict
        :type f: class:`function`
    """

    _name = f"{f.__module__.split('.')[-1]}.{f.__name__}()"
    _monitoring.update({_name: 0})

    @wraps(f)
    def wrapper(*args, **kwargs):
        error = None

        _monitoring[_name] += 1

        try:
            result = f(*args, **kwargs)
        except Exception as e:
            error = e
            result = None

        _monitoring[_name] -= 1

        if error:
            raise error

        return result
    return wrapper



import collections
import concurrent.futures
import logging
import time
from datetime import timedelta

from concurrentevents._exceptions import Cancel, StartError
from concurrentevents.event import Event, Start, Exit
from concurrentevents.eventhandler import EventHandler
from concurrentevents.tools import threadmonitor

logger = logging.getLogger('concurrentevents')


class EventManager:
    """
    A bare bones event framework using concurrent futures for threading

    Kwargs:
        :param threads: The number of threads, default to none for automatic calculation
        :type threads: int, optional
        :param class:'verbosity': The level of logs expected from the event system
        :type class:'verbosity': class:`concurrentevents.Verbosity`, int, optional
        :param process_tracking: Boolean for output of event system progress to logger
        :type process_tracking: bool, optional
        :param debug: Boolean for larger debug in the main thread
        :type debug: bool, optional
    """

    def __init__(self, threads=0, process_tracking=False, debug=False):
        """Constructor Method"""
        if threads < -1 if isinstance(threads, int) else True:
            raise ValueError(f"EventManager logger argument must be a int greater than 0 not {threads}")

        if not isinstance(process_tracking, bool):
            raise TypeError(f"EventManager process_tracking argument must be a bool not {process_tracking}")

        if not isinstance(debug, bool):
            raise TypeError(f"EventManager debug argument must be a bool not {debug}")

        # Main dictionary for all event functionality
        # Stores handlers mapped to events to allow for quick access to specific ordered functions
        # Outline:
        #   key: Event
        #   value: List of handler functions
        self.handlers = collections.defaultdict(list)

        # Futures Dictionary
        # This is a private dictionary used internally so that the main thread can manage and clean
        #   the running futures generated by the event system
        # Outline:
        #   key: Future Object
        #   value: (fn, args, kwargs)
        self.__futures = {}

        self.__thread_pool = concurrent.futures.ThreadPoolExecutor(max_workers=threads or None)

        # This boolean is set so that after firing the start event on the main thread there
        #   are no other instances of it that could be created and thus no duplicate cleaners
        self.started = False

        # This boolean is set so that any dedicated while loops not made by the Thread Pool Executor
        #   will be able to check if the program is meant to end
        self.exit = None

        # Option to enable logging of progress tracking for all events needed to run
        self.progress_tracking = process_tracking
        self.fired = 0
        self.finished = 0
        self.errors = 0
        self.canceled = 0

        # Debug boolean for main loop debug statements
        self.debug = debug

    def add_handlers(self, *event_handlers):
        """
        Establishes relationships between `EventManager` and `EventHandler` or decorated function

        Args:
            :param handlers: A list of `EventHandlers` or functions that will be added to the handler dictionary
        """
        # Parse and pull all handlers passed in
        # This enforces a strict relationship between the handler and manager
        for event_handler in event_handlers:

            # If the handler is an EventHandler
            if isinstance(event_handler, EventHandler):

                # Assign self to cls event_manager
                event_handler.__class__.event_manager = self

                # Loop through class and get catch functions
                for var in dir(event_handler):
                    func = getattr(event_handler, var)
                    if hasattr(func, 'event'):
                        self.handlers[func.event].append(func)

            # Check if function is callable and has an attribute called event
            elif callable(event_handler) and hasattr(event_handler, 'event'):
                self.handlers[event_handler.event].append(event_handler)

        return self

    def __handle(self, event):
        """
        Operates function specified to handle a specific event on that event

        This is the basis for the entire event system in terms of handling events in order
        and making sure the exit event ends the program after it runs

        Args:
            :param event: An event object to be handled
            :type event: class:`concurrentevents.Event`
        """
        event_name = event.__class__.__name__

        if isinstance(event, Exit):
            self.exit = event.timeout

        if event_name in self.handlers.keys():
            handlers = sorted(self.handlers[event_name], key=lambda h: h.priority)
            logger.info(f" {[h.__name__ for h in handlers]} handling {event.__str__()}")

            for h in handlers:
                try:
                    h(*event.args, **event.kwargs)
                except Cancel:
                    if self.progress_tracking:
                        self.canceled += 1
                    break

    def _submit(self, f, *args, **kwargs):
        """
        Interacts with the threadpool and futures dictionary to organize and track events

        This is implemented as the only methodology to put a function into the threadpool because,
        it forces all futures generated by putting something in the threadpool to be added to the
        futures list

        Parameters:
            :param f: A function object
            :type f: class:`function`
            :param `*args`: Arguments for f
            :param `**kwargs`: Keyword arguments for f
        """
        future = self.__thread_pool.submit(f, *args, **kwargs)
        self.__futures[future] = (f, args, kwargs)
        return future

    def fire(self, event):
        """
        A filter method for events that checks to see if it needs to be sent to the Thread Pool

        The purpose of this function is to validate not modify anything that
        wanted to be fired as an event.
        This intermediary step helps to eliminate anything that gets fired but
        isn't an event and any events that don't have handlers

        Args:
            :param event: An event that was fired
            :type event: Event


        :raises ValueError: Raised for when a non event is passed into the event param
        :raises KeyError: Raised for when there no handlers for the event
        """
        if not isinstance(event, Event):
            raise ValueError

        # Check to see if there are handlers ready to catch the event
        if event.__class__.__name__ not in self.handlers.keys() and not isinstance(event, Exit):
            raise KeyError(f"No Handlers For {event.__class__.__name__}")

        # DEBUG
        if not self.exit:
            logger.debug(f"Firing {event}")

            self._submit(self.__handle, event)
            if self.progress_tracking:
                self.fired += 1

    def start(self):
        """
        Starts the event sequence

        Fires the start event to trigger the beginning of all functionality
        After firing the start event the main while loop is started for debug
        and cleaning of completed futures
        """
        if self.started:
            raise StartError("EventManager is already running")
        self.started = True

        s = time.time()

        self.fire(Start())

        while not (self.exit or len(self.__futures.keys()) == 0):
            if self.debug:
                futures = len(list(self.__futures))
                running = len([ft for ft in list(self.__futures.keys()) if ft.running()])
                monitored = threadmonitor._monitoring

                process_info = f"EventManager Loop (futures={futures}, " \
                               f"running={running}, " \
                               f"monitored={monitored})"
                logger.info(process_info)

            for f, fn in self.__futures.copy().items():
                if f.done():
                    self.__futures.pop(f)
                    try:
                        if self.progress_tracking:
                            self.finished += 1

                            if fn[0].__name__ != '__handle':
                                future_info = f"{fn[0].__name__}(args={fn[1]}, kwargs={fn[2]})"
                            else:
                                future_info = fn[1][0]
                            logger.debug(f"{future_info} ({self.finished}/{self.fired})")

                        # Raises errors
                        f.result()
                    except Exception:
                        logger.exception(fn)
                        if self.progress_tracking:
                            self.errors += 1

            time.sleep(0.25)

        if not self.exit:
            self.fire(Exit())

        waited = concurrent.futures.wait(self.__futures.keys(), timeout=self.exit)
        if self.progress_tracking:
            self.finished += len(waited.done)

        if len(waited.not_done) > 0:
            logger.error("Final Cleaning Timed Out")

        logger.debug(f"Runtime {timedelta(seconds=time.time() - s)}")
        if self.progress_tracking:
            logger.debug(f"Events Fired ({self.fired})")

            logger.debug(f"Events Finished ({self.finished})")
            logger.debug(f"Events Errors ({self.errors})")

            logger.debug(f"Events Canceled ({self.canceled})")

        self.__thread_pool.shutdown()



import time
from functools import wraps


def concurrent_work_limiter(limit, timeout=0):
    """
    This decorator tool is used to limit the concurrent usage of a specific function

    Specifically any implementation keeps track of concurrent running instances of a specific function and delays
    running additional ones until the previous ones are finished

    Args:
        :param limit: A limit on how many executions of this function should happen simultaneously
        :type limit: int
    Kwargs:
        :param timeout: Timeout argument in case concurrent work takes too long
        :type timeout: int, optional

    :raises ValueError: On incorrect, limit value or timeout value
    """
    if not isinstance(limit, int):
        raise TypeError(f"concurrent_work_limiter() limit argument must be an int, not {limit}")

    if timeout < 0 if isinstance(timeout, int) else True:
        raise ValueError(f"concurrent_work_limiter() timeout argument must be an int greater than 0, not {limit}")

    def handler_function(func):
        func.working = 0
        _name = f"{func.__module__.split('.')[-1]}.{func.__name__}()"

        @wraps(func)
        def wrapper(*args, **kwargs):
            error = None

            # Delay function call till others finish
            endtime = time.monotonic() + timeout
            while func.working >= limit:
                if timeout > 0:
                    if endtime - time.monotonic() <= 0.0:
                        raise TimeoutError(f"ConcurrentWorkLimiter({_name}) timeout({timeout}s)")
                time.sleep(0.25)

            func.working += 1

            try:
                result = func(*args, **kwargs)
            except Exception as e:
                error = e
                result = None

            func.working -= 1

            if error:
                raise error

            return result
        return wrapper
    return handler_function



import time
from functools import wraps


def rate_limiter(limit, period=1):
    """
    This is an implementation of the Generic Cell Rate Algorithm which allows
    for rate limiting of function calls without the need for outside interaction

    Wikipedia: https://en.wikipedia.org/wiki/Generic_cell_rate_algorithm

    :param limit: The maximum amount of calls acceptable for the period
    :param period: The period at which the limit is applied to. Default 1
    :return:
    """

    if limit < 0 if isinstance(limit, int) else False:
        raise ValueError(f"rate_limiter() limit argument must be an int greater than 0, not {limit}")

    if period < 0 if isinstance(period, int) else False:
        raise ValueError(f"rate_limiter() period argument must be an int greater than 0, not {period}")

    # Do setup for variables for this specific decoration
    last_check = time.monotonic()
    interval = period / limit
    tat = 0

    def handlerfunction(func):
        @wraps(func)
        def wrapper(*args, **kwargs):

            # Represents the time of the call
            t = time.monotonic()

            # Theoretical Arrival Time
            # Find the maximum point either tat or now
            tat = max(last_check, t)

            # Check to see if that time is not within the window
            # If it's not within sleep until the window comes up
            if tat - t <= (period - interval):
                # Determine the new theoretical arrival time
                new_tat = max(tat, t) + interval
                # Set that tat for the next call to use
                tat = new_tat
                # Sleep for the remaining time needed to finish this window
                time.sleep(new_tat - t)

            error = None

            try:
                result = func(*args, **kwargs)
            except Exception as e:
                error = e
                result = None

            if error:
                raise error

            return result
        return wrapper
    return handlerfunction



from enum import Enum


class Priority(Enum):
    """
    Enum is used to order handlers for a specific event

    The idea is that each event is a thread so that in order
    to let certain handlers go before less important ones

    Example:

    .. highlight:: python
    .. code-block:: python

        @Catch(Start, Priority.CRITICAL)
        def foo():
            pass
    """
    CRITICAL = 0
    HIGH = 10
    MEDIUM = 20
    LOW = 30
    END = 40

    def __int__(self):
        return self.value



import logging
import queue
import time
import uuid

from concurrentevents.eventhandler import EventHandler
from concurrentevents.tools.threadmonitor import thread_monitor

logger = logging.getLogger('concurrentevents')


class ResourcePool:
    """
    Organizational tool used for resource load balancing with options for thread safety

    Uses carousel queue to balance loads between resources
    If it is specified to be thread_save additional resources are created to
    match the number of threads running and resource use it checked when
    requesting a resource

    Args:
        :param resource: An initialized resource object
        :type resource: class:`concurrentevents.Resource`
        :param amount: A number of resource to have in rotation
        :type amount: int
    Kwargs
        :param thread_safe: An option to require all resource and use to be thread safe
        :type thread_safe: bool, optional
    """
    def __init__(self, resource, amount, thread_safe=False):
        """Constructor Method"""
        try:
            if not (isinstance(resource, Resource) or issubclass(resource.__class__, Resource)):
                raise TypeError
        except TypeError:
            raise TypeError(
                f"ResourceManager() resource argument must be a "
                f"Resource or subclass of Resource, not {resource}")

        if not isinstance(amount, int):
            raise TypeError(f"ResourceManager() amount argument must be an int, not {amount}")

        if not isinstance(thread_safe, bool):
            raise TypeError(f"ResourceManager() amount argument must be a bool, not {thread_safe}")

        self.thread_safe = thread_safe

        self.__resources = queue.Queue()

        # Loop to copy a resource to populate __resources
        self.__resources.put(resource)
        for i in range(1, amount):
            self.__resources.put(resource.copy())

    @thread_monitor
    def get(self, timeout=0):
        """
        Returns a resource object from the queue of resources with a timeout

        Args:
            :param timeout: A timeout variable set to none to be used in the queue.get(timeout) function
            :type timeout: int, None, optional

        :return: One of the resource from the pool
        :rtype: class:`concurrentevents.Resource`
        """
        if not isinstance(timeout, int) or timeout < 0:
            raise ValueError("timeout must be an int greater than zero")

        res = self.__resources.get(timeout=timeout)
        self.__resources.put(res, timeout=timeout)

        if self.thread_safe:
            endtime = time.monotonic() + timeout
            while res._in_use:
                res = self.__resources.get(timeout=timeout)
                self.__resources.put(res, timeout=timeout)

                if timeout:
                    if endtime - time.monotonic() <= 0.0:
                        raise TimeoutError

                time.sleep(0.25)

        res._in_use = True
        return res


class Resource(EventHandler):
    """
    A generic class to wrap a function or class and allow it to be used in a resource pool

    Args:
        :param f: A function object to be pasted in for executions
        :type f: class:`function`
    """
    def __init__(self, f=None, **kwargs):
        """Constructor Method"""
        self._f = f
        self.id = uuid.uuid4()
        self._in_use = False
        self._kwargs = kwargs

        super().__init__(**kwargs)

    def copy(self):
        """
        A copy constructor to allow for duplicates to be made to populate the resource pool

        :return: A new instance of the same class
        """
        try:
            if issubclass(self.__class__, Resource) or isinstance(self, Resource):
                return self.__class__(self._f or None, **self._kwargs)
        except TypeError:
            return self.__class__()

    def __call__(self, *args, **kwargs):
        """
        Executes the stored function with args and keyword args

        Args:
            :param args: Arguments for the internal function
            :param kwargs: Keyword Arguments for the internal function

        :return: Returns any results from the function
        """
        return self._f(*args, **kwargs)

    def open(self):
        """
        Empty function to override if functionality wants to be
        added to the __enter__ operation
        """

    def close(self):
        """
        Function can either be override to add function to the with statement or
        if no with statement is used this signals that the object is no longer
        in use __exit__ operation
        """
        self._in_use = False

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._in_use = False
        self.close()
        if exc_type:
            logger.exception(exc_type)
            return False
        return True



class Event:
    """
    Basic Event Class

    This class is meant to be used as a superclass for specific event
    Those classes than inherit the Event class can now be used with @Catch(PrintEvent)
    and when fired to the event manager with handlers catching it will be run

    .. highlight:: python
    .. code-block:: python

        class PrintEvent(Event):
            def __init__(self, toprint)
                Event.__init__(self, toprint=toprint)

        class PrintHandler(EventHandler):
            @Catch(PrintEvent)
            def printAnything(toprint)
                print(toprint)

    Args:
        :param `*args`: Arguments for handling functions
        :param `**kwargs`: Keyword arguments for handling function
    """
    def __init__(self, *args, **kwargs):
        """Constructor Method"""
        self.args = args
        self.kwargs = kwargs

    def __str__(self):
        """
        Created a string of an event

        >>>str(Start)
        'Start(args=(foo), kwargs={'foo':bar})'

        :return: A string of the event including args and kwargs
        :rtype: str
        """
        return f"{self.__class__.__name__}(args={self.args}, kwargs={self.kwargs})"


class Start(Event):
    """An implementation of the Event class to signal program start"""


class Exit(Event):
    """
    An implementation of the Event class to signal program exit

    Args:
        :param timeout: A timeout option for shutdown operations
        :type timeout: int, optional
    """
    def __init__(self, timeout=5):
        """Constructor Method"""
        self.timeout = timeout
        super().__init__()



import setuptools

with open("README.rst", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="ConcurrentEvents",
    version="1.11.2",
    author="Reggles",
    author_email="reginaldbeakes@gmail.com",
    description="An event system build on top of the concurrent futures library, including additional threading tools",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitlab.com/Reggles44/concurrentevents",
    include_package_data=True,
    packages=['concurrentevents', 'concurrentevents.enums', 'concurrentevents.tools'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ]
)



from concurrentevents._exceptions import Cancel


class EventHandler:
    """
    The main class used to implement a function to handle an event

    This is not required to handle function but rather standardizes useful
    functionality to simplify implementation

    Args:
        :param `**kwargs`: Keyword arguments used to set member variables
    """
    event_manager = None

    def __init__(self, **kwargs):
        """Constructor Method"""
        for keyword, arg in kwargs.items():
            setattr(self, keyword, arg)

    @classmethod
    def fire(cls, event):
        """
        Wrapper function for the event manager fire function

        Allows for function under an implemented event handler class to fire new events

        Args:
            :param event: A event to be fired through the event manager
            :type event: class:`concurrentevents.Event`
        """
        cls.event_manager.fire(event)

    @staticmethod
    def cancel():
        """
        A shortcut function to raise the cancel error signaling the __handler function to stop
        :raises Cancel: This error is a shortcut to trigger changes in the handling of an event
        """
        raise Cancel()



# Core
# Exceptions
from concurrentevents._exceptions import EventError, Cancel
from concurrentevents.catch import catch
# Enums
from concurrentevents.enums import Priority
from concurrentevents.event import Event, Start, Exit
from concurrentevents.eventhandler import EventHandler
from concurrentevents.eventmanager import EventManager
# Tools
from concurrentevents.tools.concurrentworklimiter import concurrent_work_limiter
from concurrentevents.tools.ratelimiter import rate_limiter
from concurrentevents.tools.resourcepool import ResourcePool, Resource
from concurrentevents.tools.threadmonitor import thread_monitor



from functools import wraps

from concurrentevents.enums import Priority
from concurrentevents.event import Event


def catch(event, priority=Priority.MEDIUM):
    """
    This is a decorator for adding member variables to a function
    that should be a handler of a specific event

    .. highlight:: python
    .. code-block:: python

        class ExampleHandler(EventHandler):

            @Catch(Start)
            def handleStart():
                foo()

            @Catch("Start")
            def handleStartFromString():
                bar()

            @Catch(Exit, priority=Priority.CRITICAL)
            def handleExitWithCriticalPriority():
                pass

    Args:
        :param event: Takes in a representation of an event
        :type event: str, class:`concurrentevents.Event`
    Kwargs:
        :param priority: An argument used to signal order to handlers for a specific event
        :type priority: class:`concurrentevents.Priority`, optional

    :raises TypeError: If any unaccepted values are passed in for either argument
    """
    try:
        if isinstance(event, str):
            evt = event
        elif issubclass(event, Event):
            evt = event.__name__
        else:
            raise TypeError
    except TypeError:
        raise TypeError(f"Catch() event argument must be an event or string, not {event}")

    try:
        prt = int(priority)
    except TypeError:
        raise TypeError(f"Catch() priority argument must be convertible to an int, not {priority}")

    def decorator(func):
        func.event = evt
        func.priority = prt

        @wraps(func)
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        return wrapper
    return decorator
